import contextlib
import io
import textwrap

import mock
import pytest

from uaclient import exceptions, messages
from uaclient.api.u.pro.security.fix._common import FixStatus, UnfixedPackage
from uaclient.api.u.pro.security.fix._common.plan.v1 import (
    ESM_APPS_POCKET,
    ESM_INFRA_POCKET,
    STANDARD_UPDATES_POCKET,
    AptUpgradeData,
    AttachData,
    EnableData,
    FailUpdatingESMCacheData,
    FixPlanAptUpgradeStep,
    FixPlanAttachStep,
    FixPlanEnableStep,
    FixPlanNoOpLivepatchFixStep,
    FixPlanNoOpStatus,
    FixPlanNoOpStep,
    FixPlanResult,
    FixPlanUSNResult,
    FixPlanWarningFailUpdatingESMCache,
    FixPlanWarningPackageCannotBeInstalled,
    FixPlanWarningSecurityIssueNotFixed,
    NoOpData,
    NoOpLivepatchFixData,
    PackageCannotBeInstalledData,
    SecurityIssueNotFixedData,
    USNAdditionalData,
)
from uaclient.api.u.pro.security.fix.usn.plan.v1 import (
    USNFixPlanResult,
    USNSFixPlanResult,
)
from uaclient.cli.fix import (
    FixContext,
    _execute_apt_upgrade_step,
    _execute_attach_step,
    _execute_enable_step,
    _handle_subscription_for_required_service,
    _perform_magic_attach,
    action_fix,
    execute_fix_plan,
    fix_usn,
    print_cve_header,
    print_usn_header,
)
from uaclient.entitlements.entitlement_status import (
    ApplicabilityStatus,
    UserFacingStatus,
)
from uaclient.files.notices import Notice
from uaclient.status import colorize_commands

M_PATH = "uaclient.cli.fix."


class TestActionFix:
    @pytest.mark.parametrize(
        "issue,is_valid",
        (
            ("CVE-2020-1234", True),
            ("cve-2020-12345", True),
            ("cve-1234-123456", True),
            ("CVE-2020-1234567", True),
            ("USN-1234-1", True),
            ("usn-1234-12", True),
            ("USN-12345-1", True),
            ("usn-12345-12", True),
            ("lsn-1234-1", True),
            ("LSN-1234-12", True),
            ("LSN-1234-123", False),
            ("cve-1234-123", False),
            ("CVE-1234-12345678", False),
            ("USA-1234-12345678", False),
        ),
    )
    @mock.patch("uaclient.cli.fix.fix_cve")
    @mock.patch("uaclient.cli.fix.fix_usn")
    def test_attached(self, m_fix_cve, m_fix_usn, issue, is_valid, FakeConfig):
        """Check that root and non-root will emit attached status"""
        cfg = FakeConfig()
        args = mock.MagicMock(
            security_issue=issue, dry_run=False, no_related=False
        )
        m_fix_cve.return_value = FixStatus.SYSTEM_NON_VULNERABLE
        m_fix_usn.return_value = FixStatus.SYSTEM_NON_VULNERABLE
        if is_valid:
            assert 0 == action_fix(args, cfg=cfg)
        else:
            with pytest.raises(exceptions.UbuntuProError) as excinfo:
                action_fix(args, cfg=cfg)

            expected_msg = (
                'Error: issue "{}" is not recognized.\n'
                'Usage: "pro fix CVE-yyyy-nnnn" or "pro fix USN-nnnn"'
            ).format(issue)

            assert expected_msg == str(excinfo.value)
            assert 0 == m_fix_cve.call_count
            assert 0 == m_fix_usn.call_count

    def test_cve_header(self):
        cve = FixPlanResult(
            title="CVE-2020-1472",
            description="Samba vulnerability",
            current_status="not-affected",
            expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
            affected_packages=["pkg1", "pkg2"],
            plan=[],
            warnings=None,
            error=None,
            additional_data=None,
        )

        fake_stdout = io.StringIO()
        with contextlib.redirect_stdout(fake_stdout):
            print_cve_header(cve)

        assert (
            textwrap.dedent(
                """\
                CVE-2020-1472: Samba vulnerability
                 - https://ubuntu.com/security/CVE-2020-1472"""
            )
            == fake_stdout.getvalue().strip()
        )

    @pytest.mark.parametrize(
        "usn,expected_output",
        (
            (
                FixPlanUSNResult(
                    target_usn_plan=FixPlanResult(
                        title="USN-4510-2",
                        current_status="not-affected",
                        description="Samba vulnerability",
                        expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,  # noqa
                        affected_packages=["pkg1", "pkg2"],
                        plan=[],
                        warnings=None,
                        error=None,
                        additional_data=USNAdditionalData(
                            associated_cves=["CVE-2020-1473", "CVE-2020-1472"],
                            associated_launchpad_bugs=[],
                        ),
                    ),
                    related_usns_plan=[],
                ),
                textwrap.dedent(
                    """\
                    USN-4510-2: Samba vulnerability
                    Associated CVEs:
                     - https://ubuntu.com/security/CVE-2020-1473
                     - https://ubuntu.com/security/CVE-2020-1472"""
                ),
            ),
            (
                FixPlanUSNResult(
                    target_usn_plan=FixPlanResult(
                        title="USN-4038-3",
                        current_status="not-affected",
                        description="USN vulnerability",
                        expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,  # noqa
                        affected_packages=["pkg1", "pkg2"],
                        plan=[],
                        warnings=None,
                        error=None,
                        additional_data=USNAdditionalData(
                            associated_cves=[],
                            associated_launchpad_bugs=[
                                "https://launchpad.net/bugs/1834494"
                            ],
                        ),
                    ),
                    related_usns_plan=[],
                ),
                textwrap.dedent(
                    """\
                    USN-4038-3: USN vulnerability
                    Found Launchpad bugs:
                     - https://launchpad.net/bugs/1834494"""
                ),
            ),
        ),
    )
    def test_usn_header(self, usn, expected_output):
        fake_stdout = io.StringIO()
        with contextlib.redirect_stdout(fake_stdout):
            print_usn_header(usn)
        assert expected_output == fake_stdout.getvalue().strip()


class TestExecuteFixPlan:
    @pytest.mark.parametrize(
        "fix_plan,dry_run,cloud_type,expected_output,"
        "expected_fix_status,expected_unfixed_pkgs",
        (
            (  # No affected_packages listed
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="not-affected",
                    expected_status=FixStatus.SYSTEM_NOT_AFFECTED.value.msg,
                    affected_packages=[],
                    plan=[
                        FixPlanNoOpStep(
                            data=NoOpData(
                                status=FixPlanNoOpStatus.NOT_AFFECTED.value
                            ),
                            order=1,
                        )
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                False,
                (None, None),
                textwrap.dedent(
                    """\
                   No affected source packages are installed.

                   {check} USN-### does not affect your system.
                   """.format(
                        check=messages.OKGREEN_CHECK
                    )
                ),
                FixStatus.SYSTEM_NOT_AFFECTED,
                [],
            ),
            (  # CVE already fixed by Livepatch
                FixPlanResult(
                    title="CVE-###",
                    description="test",
                    current_status="fixed",
                    expected_status=FixStatus.SYSTEM_NOT_AFFECTED.value.msg,
                    affected_packages=[],
                    plan=[
                        FixPlanNoOpLivepatchFixStep(
                            data=NoOpLivepatchFixData(
                                status="cve-fixed-by-livepatch",
                                patch_version="87.1",
                            ),
                            order=1,
                        )
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                False,
                (None, None),
                messages.CVE_FIXED_BY_LIVEPATCH.format(
                    issue="CVE-###",
                    version="87.1",
                ),
                FixStatus.SYSTEM_NON_VULNERABLE,
                [],
            ),
            (  # version is >= released affected package
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["slsrc"],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=[],
                                source_packages=["slsrc"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=1,
                        )
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                False,
                (None, None),
                textwrap.dedent(
                    """\
                   1 affected source package is installed: slsrc
                   (1/1) slsrc:
                   A fix is available in Ubuntu standard updates.
                   The update is already installed.

                   {check} USN-### is resolved.
                   """.format(
                        check=messages.OKGREEN_CHECK
                    )
                ),
                FixStatus.SYSTEM_NON_VULNERABLE,
                [],
            ),
            (  # installing package fix
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["slsrc"],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["sl"],
                                source_packages=["slsrc"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=1,
                        )
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                False,
                (None, None),
                textwrap.dedent(
                    """\
                   1 affected source package is installed: slsrc
                   (1/1) slsrc:
                   A fix is available in Ubuntu standard updates.
                   """
                )
                + colorize_commands(
                    [["apt update && apt install --only-upgrade" " -y sl"]]
                )
                + "\n\n"
                + "{check} USN-### is resolved.\n".format(
                    check=messages.OKGREEN_CHECK
                ),
                FixStatus.SYSTEM_NON_VULNERABLE,
                [],
            ),
            (  # installing package fix that comes from esm-infra
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["slsrc"],
                    plan=[
                        FixPlanAttachStep(
                            data=AttachData(
                                reason="test",
                                required_service=ESM_INFRA_POCKET,
                                source_packages=["slsrc"],
                            ),
                            order=1,
                        ),
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_INFRA_POCKET,
                                source_packages=["slsrc"],
                            ),
                            order=2,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["sl"],
                                source_packages=["slsrc"],
                                pocket=ESM_INFRA_POCKET,
                            ),
                            order=3,
                        ),
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                False,
                ("azure", None),
                textwrap.dedent(
                    """\
                    1 affected source package is installed: slsrc
                    (1/1) slsrc:
                    A fix is available in Ubuntu Pro: ESM Infra.
                    """
                )
                + "\n".join(
                    [
                        messages.SECURITY_USE_PRO_TMPL.format(
                            title="Azure",
                            cloud_specific_url="https://ubuntu.com/azure/pro",
                        ),
                        messages.SECURITY_UPDATE_NOT_INSTALLED_SUBSCRIPTION,
                    ]
                ),
                FixStatus.SYSTEM_STILL_VULNERABLE,
                [
                    UnfixedPackage(
                        pkg="slsrc",
                        unfixed_reason=messages.SECURITY_UA_SERVICE_REQUIRED.format(  # noqa
                            service="esm-infra"
                        ),
                    )
                ],
            ),
            (  # installing package fixes that comes from
                # standard-updates and esm-infra
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["curl", "slsrc"],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["curl"],
                                source_packages=["curl"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=1,
                        ),
                        FixPlanAttachStep(
                            data=AttachData(
                                reason="test",
                                required_service=ESM_APPS_POCKET,
                                source_packages=["slsrc"],
                            ),
                            order=2,
                        ),
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_APPS_POCKET,
                                source_packages=["slsrc"],
                            ),
                            order=3,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["sl"],
                                source_packages=["slsrc"],
                                pocket=ESM_APPS_POCKET,
                            ),
                            order=4,
                        ),
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                False,
                ("gce", None),
                textwrap.dedent(
                    """\
                    2 affected source packages are installed: curl, slsrc
                    (1/2) curl:
                    A fix is available in Ubuntu standard updates.
                    """
                )
                + colorize_commands(
                    [["apt update && apt install --only-upgrade" " -y curl"]]
                )
                + "\n"
                + textwrap.dedent(
                    """\
                    (2/2) slsrc:
                    A fix is available in Ubuntu Pro: ESM Apps.
                    """
                )
                + "\n".join(
                    [
                        messages.SECURITY_USE_PRO_TMPL.format(
                            title="GCP",
                            cloud_specific_url="https://ubuntu.com/gcp/pro",
                        ),
                        messages.SECURITY_UPDATE_NOT_INSTALLED_SUBSCRIPTION,
                    ]
                )
                + "\n\n"
                + "1 package is still affected: slsrc",
                FixStatus.SYSTEM_STILL_VULNERABLE,
                [
                    UnfixedPackage(
                        pkg="slsrc",
                        unfixed_reason=messages.SECURITY_UA_SERVICE_REQUIRED.format(  # noqa
                            service="esm-apps"
                        ),
                    )
                ],
            ),
            (  # installing package fix that are not yet available
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=[
                        "pkg1",
                        "pkg2",
                        "pkg3",
                        "pkg4",
                        "pkg5",
                        "pkg6",
                        "pkg7",
                        "pkg8",
                        "pkg9",
                        "pkg10",
                        "pkg11",
                        "pkg12",
                        "pkg13",
                        "pkg14",
                        "pkg15",
                    ],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg10", "pkg11"],
                                source_packages=["pkg10", "pkg11"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=5,
                        ),
                        FixPlanAttachStep(
                            data=AttachData(
                                reason="test",
                                required_service=ESM_INFRA_POCKET,
                                source_packages=["pkg12", "pkg13"],
                            ),
                            order=6,
                        ),
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_INFRA_POCKET,
                                source_packages=["pkg12", "pkg13"],
                            ),
                            order=7,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg12", "pkg13"],
                                source_packages=["pkg12", "pkg13"],
                                pocket=ESM_INFRA_POCKET,
                            ),
                            order=8,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg14", "pkg15"],
                                source_packages=["pkg14", "pkg15"],
                                pocket=ESM_APPS_POCKET,
                            ),
                            order=9,
                        ),
                    ],
                    warnings=[
                        FixPlanWarningSecurityIssueNotFixed(
                            data=SecurityIssueNotFixedData(
                                source_packages=["pkg1", "pkg2", "pkg9"],
                                status="ignored",
                            ),
                            order=1,
                        ),
                        FixPlanWarningSecurityIssueNotFixed(
                            data=SecurityIssueNotFixedData(
                                source_packages=["pkg7", "pkg8"],
                                status="needed",
                            ),
                            order=2,
                        ),
                        FixPlanWarningSecurityIssueNotFixed(
                            data=SecurityIssueNotFixedData(
                                source_packages=["pkg5", "pkg6"],
                                status="needs-triage",
                            ),
                            order=3,
                        ),
                        FixPlanWarningSecurityIssueNotFixed(
                            data=SecurityIssueNotFixedData(
                                source_packages=["pkg3", "pkg4"],
                                status="pending",
                            ),
                            order=4,
                        ),
                    ],
                    error=None,
                    additional_data=None,
                ),
                False,
                ("gce", None),
                textwrap.dedent(
                    """\
                    15 affected source packages are installed: {}
                    (1/15, 2/15, 3/15) pkg1, pkg2, pkg9:
                    Sorry, no fix is available.
                    (4/15, 5/15) pkg7, pkg8:
                    Sorry, no fix is available yet.
                    (6/15, 7/15) pkg5, pkg6:
                    Ubuntu security engineers are investigating this issue.
                    (8/15, 9/15) pkg3, pkg4:
                    A fix is coming soon. Try again tomorrow.
                    (10/15, 11/15) pkg10, pkg11:
                    A fix is available in Ubuntu standard updates.
                    """
                ).format(
                    (
                        "pkg1, pkg10, pkg11, pkg12, pkg13,\n"
                        "    pkg14, pkg15, pkg2, pkg3, pkg4, pkg5, pkg6, pkg7,"
                        " pkg8, pkg9"
                    )
                )
                + colorize_commands(
                    [
                        [
                            "apt update && apt install --only-upgrade"
                            " -y pkg10 pkg11"
                        ]
                    ]
                )
                + "\n"
                + textwrap.dedent(
                    """\
                    (12/15, 13/15) pkg12, pkg13:
                    A fix is available in Ubuntu Pro: ESM Infra.
                    """
                )
                + "\n".join(
                    [
                        messages.SECURITY_USE_PRO_TMPL.format(
                            title="GCP",
                            cloud_specific_url="https://ubuntu.com/gcp/pro",
                        ),
                        messages.SECURITY_UPDATE_NOT_INSTALLED_SUBSCRIPTION,
                    ]
                )
                + "\n\n"
                + "11 packages are still affected: {}".format(
                    (
                        "pkg1, pkg12, pkg13, pkg2, pkg3, pkg4, pkg5,\n"
                        "    pkg6, pkg7, pkg8, pkg9"
                    )
                ),
                FixStatus.SYSTEM_STILL_VULNERABLE,
                [
                    UnfixedPackage(
                        pkg="pkg1",
                        unfixed_reason=messages.SECURITY_CVE_STATUS_IGNORED,
                    ),
                    UnfixedPackage(
                        pkg="pkg2",
                        unfixed_reason=messages.SECURITY_CVE_STATUS_IGNORED,
                    ),
                    UnfixedPackage(
                        pkg="pkg9",
                        unfixed_reason=messages.SECURITY_CVE_STATUS_IGNORED,
                    ),
                    UnfixedPackage(
                        pkg="pkg7",
                        unfixed_reason=messages.SECURITY_CVE_STATUS_NEEDED,
                    ),
                    UnfixedPackage(
                        pkg="pkg8",
                        unfixed_reason=messages.SECURITY_CVE_STATUS_NEEDED,
                    ),
                    UnfixedPackage(
                        pkg="pkg5",
                        unfixed_reason=messages.SECURITY_CVE_STATUS_TRIAGE,
                    ),
                    UnfixedPackage(
                        pkg="pkg6",
                        unfixed_reason=messages.SECURITY_CVE_STATUS_TRIAGE,
                    ),
                    UnfixedPackage(
                        pkg="pkg3",
                        unfixed_reason=messages.SECURITY_CVE_STATUS_PENDING,
                    ),
                    UnfixedPackage(
                        pkg="pkg4",
                        unfixed_reason=messages.SECURITY_CVE_STATUS_PENDING,
                    ),
                    UnfixedPackage(
                        pkg="pkg12",
                        unfixed_reason=messages.SECURITY_UA_SERVICE_REQUIRED.format(  # noqa
                            service="esm-infra"
                        ),
                    ),
                    UnfixedPackage(
                        pkg="pkg13",
                        unfixed_reason=messages.SECURITY_UA_SERVICE_REQUIRED.format(  # noqa
                            service="esm-infra"
                        ),
                    ),
                ],
            ),
            (  # installing package fix that are not yet available
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="not-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=[
                        "longpackagename1",
                        "longpackagename2",
                        "longpackagename3",
                        "longpackagename4",
                        "longpackagename5",
                    ],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=[
                                    "longpackagename1",
                                    "longpackagename2",
                                    "longpackagename3",
                                    "longpackagename4",
                                    "longpackagename5",
                                ],
                                source_packages=[
                                    "longpackagename1",
                                    "longpackagename2",
                                    "longpackagename3",
                                    "longpackagename4",
                                    "longpackagename5",
                                ],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=1,
                        ),
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                False,
                ("gce", None),
                """\
5 affected source packages are installed: longpackagename1, longpackagename2,
    longpackagename3, longpackagename4, longpackagename5
(1/5, 2/5, 3/5, 4/5, 5/5) longpackagename1, longpackagename2, longpackagename3,
    longpackagename4, longpackagename5:
A fix is available in Ubuntu standard updates.\n"""
                + colorize_commands(
                    [
                        [
                            "apt update && apt install --only-upgrade"
                            " -y longpackagename1 longpackagename2 "
                            "longpackagename3 longpackagename4 "
                            "longpackagename5"
                        ]
                    ]
                )
                + "\n\n"
                + "{check} USN-### is resolved.\n".format(
                    check=messages.OKGREEN_CHECK
                ),
                FixStatus.SYSTEM_NON_VULNERABLE,
                [],
            ),
            (
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_STILL_VULNERABLE.value.msg,  # noqa
                    affected_packages=["pkg1", "pkg2"],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg2"],
                                source_packages=["pkg2"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=2,
                        ),
                    ],
                    warnings=[
                        FixPlanWarningPackageCannotBeInstalled(
                            data=PackageCannotBeInstalledData(
                                binary_package="pkg1",
                                binary_package_version="2.0",
                                source_package="pkg1",
                                related_source_packages=["pkg1", "pkg2"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=1,
                        )
                    ],
                    error=None,
                    additional_data=None,
                ),
                False,
                (None, None),
                textwrap.dedent(
                    """\
                    2 affected source packages are installed: pkg1, pkg2
                    (1/2, 2/2) pkg1, pkg2:
                    A fix is available in Ubuntu standard updates.
                    - Cannot install package pkg1 version 2.0
                    """
                )
                + colorize_commands(
                    [["apt update && apt install --only-upgrade" " -y pkg2"]]
                )
                + "\n\n"
                + "1 package is still affected: pkg1"
                + "\n"
                + "{check} USN-### is not resolved.\n".format(
                    check=messages.FAIL_X
                ),
                FixStatus.SYSTEM_STILL_VULNERABLE,
                [
                    UnfixedPackage(
                        pkg="pkg1",
                        unfixed_reason="Cannot install package pkg1 version 2.0",  # noqa
                    ),
                ],
            ),
            (
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_STILL_VULNERABLE.value.msg,  # noqa
                    affected_packages=["pkg1", "pkg2"],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg2"],
                                source_packages=["pkg2"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=3,
                        ),
                    ],
                    warnings=[
                        FixPlanWarningFailUpdatingESMCache(
                            data=FailUpdatingESMCacheData(
                                title="Error msg",
                                code="error-code",
                            ),
                            order=1,
                        ),
                        FixPlanWarningPackageCannotBeInstalled(
                            data=PackageCannotBeInstalledData(
                                binary_package="pkg1",
                                binary_package_version="2.0",
                                source_package="pkg1",
                                related_source_packages=["pkg1", "pkg2"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=2,
                        ),
                    ],
                    error=None,
                    additional_data=None,
                ),
                False,
                (None, None),
                textwrap.dedent(
                    """\
                    2 affected source packages are installed: pkg1, pkg2
                    WARNING: Failed to update ESM cache - package availability may be inaccurate
                    (1/2, 2/2) pkg1, pkg2:
                    A fix is available in Ubuntu standard updates.
                    - Cannot install package pkg1 version 2.0
                    """  # noqa
                )
                + colorize_commands(
                    [["apt update && apt install --only-upgrade" " -y pkg2"]]
                )
                + "\n\n"
                + "1 package is still affected: pkg1"
                + "\n"
                + "{check} USN-### is not resolved.\n".format(
                    check=messages.FAIL_X
                ),
                FixStatus.SYSTEM_STILL_VULNERABLE,
                [
                    UnfixedPackage(
                        pkg="pkg1",
                        unfixed_reason="Cannot install package pkg1 version 2.0",  # noqa
                    ),
                ],
            ),
        ),
    )
    @mock.patch("uaclient.system.should_reboot", return_value=False)
    @mock.patch("uaclient.cli.fix.get_cloud_type")
    @mock.patch("uaclient.util.prompt_choices", return_value="c")
    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.apt.run_apt_command")
    def test_execute_fix_plan(
        self,
        _m_apt_run_cmd,
        _m_apt_run_update,
        _m_prompt,
        m_get_cloud_type,
        _m_should_reboot,
        fix_plan,
        dry_run,
        cloud_type,
        expected_output,
        expected_fix_status,
        expected_unfixed_pkgs,
        capsys,
        FakeConfig,
    ):
        m_get_cloud_type.return_value = cloud_type
        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            assert (
                expected_fix_status,
                expected_unfixed_pkgs,
            ) == execute_fix_plan(fix_plan, dry_run, cfg=FakeConfig())

        out, _ = capsys.readouterr()
        assert expected_output in out

    @pytest.mark.parametrize(
        "fix_plan,expected_output,expected_fix_status,expected_unfixed_pkgs",
        (
            (
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1", "pkg2", "pkg3"],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg1", "pkg2", "pkg3"],
                                source_packages=["pkg1", "pkg2", "pkg3"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=1,
                        ),
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                textwrap.dedent(
                    """\
                    3 affected source packages are installed: pkg1, pkg2, pkg3
                    (1/3, 2/3, 3/3) pkg1, pkg2, pkg3:
                    A fix is available in Ubuntu standard updates.
                    """
                )
                + colorize_commands(
                    [
                        [
                            "apt update && apt install --only-upgrade"
                            " -y pkg1 pkg2 pkg3"
                        ]
                    ]
                )
                + "\n"
                + "test exception"
                + "\n\n"
                + "3 packages are still affected: pkg1, pkg2, pkg3"
                + "\n"
                + "{check} USN-### is not resolved.\n".format(
                    check=messages.FAIL_X
                ),
                FixStatus.SYSTEM_STILL_VULNERABLE,
                [
                    UnfixedPackage(
                        pkg="pkg1",
                        unfixed_reason="test exception",
                    ),
                    UnfixedPackage(
                        pkg="pkg2",
                        unfixed_reason="test exception",
                    ),
                    UnfixedPackage(
                        pkg="pkg3",
                        unfixed_reason="test exception",
                    ),
                ],
            ),
        ),
    )
    @mock.patch("uaclient.system.should_reboot", return_value=False)
    @mock.patch("uaclient.apt.run_apt_update_command")
    def test_execute_fix_plan_apt_upgrade_fail(
        self,
        m_apt_update,
        _m_should_reboot,
        fix_plan,
        expected_output,
        expected_fix_status,
        expected_unfixed_pkgs,
        capsys,
        FakeConfig,
    ):
        m_apt_update.side_effect = Exception("test exception")
        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            assert (
                expected_fix_status,
                expected_unfixed_pkgs,
            ) == execute_fix_plan(fix_plan, dry_run=False, cfg=FakeConfig())

        out, _ = capsys.readouterr()
        assert expected_output in out

    @pytest.mark.parametrize(
        "fix_plan,expected_output,expected_fix_status,"
        "expected_unfixed_pkgs",
        (
            (  # installing package fixes that comes from
                # standard-updates and esm-infra when attaching
                # the system
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1", "pkg2", "pkg3"],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg2"],
                                source_packages=["pkg2"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=1,
                        ),
                        FixPlanAttachStep(
                            data=AttachData(
                                reason="test",
                                required_service=ESM_INFRA_POCKET,
                                source_packages=["pkg3"],
                            ),
                            order=2,
                        ),
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_INFRA_POCKET,
                                source_packages=["pkg3"],
                            ),
                            order=3,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg3"],
                                source_packages=["pkg3"],
                                pocket=ESM_INFRA_POCKET,
                            ),
                            order=4,
                        ),
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_APPS_POCKET,
                                source_packages=["pkg1"],
                            ),
                            order=5,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg1"],
                                source_packages=["pkg1"],
                                pocket=ESM_APPS_POCKET,
                            ),
                            order=6,
                        ),
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                textwrap.dedent(
                    """\
                    3 affected source packages are installed: pkg1, pkg2, pkg3
                    (1/3) pkg2:
                    A fix is available in Ubuntu standard updates.
                    """
                )
                + colorize_commands(
                    [["apt update && apt install --only-upgrade" " -y pkg2"]]
                )
                + "\n"
                + textwrap.dedent(
                    """\
                    (2/3) pkg3:
                    A fix is available in Ubuntu Pro: ESM Infra.
                    """
                )
                + messages.SECURITY_UPDATE_NOT_INSTALLED_SUBSCRIPTION
                + "\n"
                + messages.PROMPT_ENTER_TOKEN
                + "\n"
                + colorize_commands([["pro attach pro_token"]])
                + "\n"
                + colorize_commands(
                    [["apt update && apt install --only-upgrade" " -y pkg3"]]
                )
                + "\n"
                + textwrap.dedent(
                    """\
                    (3/3) pkg1:
                    A fix is available in Ubuntu Pro: ESM Apps.
                    """
                )
                + colorize_commands(
                    [["apt update && apt install --only-upgrade" " -y pkg1"]]
                )
                + "\n\n"
                + "{check} USN-### is resolved.\n".format(
                    check=messages.OKGREEN_CHECK
                ),
                FixStatus.SYSTEM_NON_VULNERABLE,
                [],
            ),
        ),
    )
    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.apt.run_apt_command")
    @mock.patch("uaclient.system.should_reboot", return_value=False)
    @mock.patch("uaclient.util.prompt_choices", return_value="a")
    @mock.patch("uaclient.cli.fix.get_cloud_type", return_value=(None, None))
    @mock.patch("builtins.input", return_value="pro_token")
    @mock.patch(M_PATH + "_handle_subscription_for_required_service")
    @mock.patch("uaclient.cli.fix.attach_with_token")
    def test_execute_fix_plan_when_attach_is_needed(
        self,
        m_attach_with_token,
        m_handle_required_service,
        _m_input,
        _m_get_cloud_type,
        _m_prompt,
        _m_should_reboot,
        _m_run_apt_command,
        _m_run_apt_update,
        fix_plan,
        expected_output,
        expected_fix_status,
        expected_unfixed_pkgs,
        fake_machine_token_file,
        capsys,
    ):
        def fake_attach(cfg, token, allow_enable):
            fake_machine_token_file.attached = True
            return 0

        m_attach_with_token.side_effect = fake_attach
        m_handle_required_service.return_value = True
        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            assert (
                expected_fix_status,
                expected_unfixed_pkgs,
            ) == execute_fix_plan(fix_plan, dry_run=False, cfg=None)

        out, _ = capsys.readouterr()
        assert expected_output in out

    @pytest.mark.parametrize(
        "service_status",
        (
            (UserFacingStatus.INACTIVE),
            (UserFacingStatus.INAPPLICABLE),
            (UserFacingStatus.UNAVAILABLE),
        ),
    )
    @pytest.mark.parametrize(
        "fix_plan,expected_output,expected_fix_status,"
        "expected_unfixed_pkgs",
        (
            (  # installing package fixes that comes from
                # standard-updates and esm-infra when attaching
                # the system
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1"],
                    plan=[
                        FixPlanAttachStep(
                            data=AttachData(
                                reason="test",
                                required_service=ESM_INFRA_POCKET,
                                source_packages=["pkg1"],
                            ),
                            order=1,
                        ),
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_INFRA_POCKET,
                                source_packages=["pkg1"],
                            ),
                            order=2,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg1"],
                                source_packages=["pkg1"],
                                pocket=ESM_INFRA_POCKET,
                            ),
                            order=3,
                        ),
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                textwrap.dedent(
                    """\
                    1 affected source package is installed: pkg1
                    (1/1) pkg1:
                    A fix is available in Ubuntu Pro: ESM Infra.
                    """
                )
                + messages.SECURITY_UPDATE_NOT_INSTALLED_SUBSCRIPTION
                + "\n"
                + messages.PROMPT_ENTER_TOKEN
                + "\n"
                + colorize_commands([["pro attach pro_token"]])
                + "\n"
                + messages.SECURITY_UA_SERVICE_NOT_ENTITLED.format(
                    service="esm-infra"
                )  # noqa
                + "\n\n"
                + "1 package is still affected: pkg1"
                + "\n"
                + "{check} USN-### is not resolved.\n".format(
                    check=messages.FAIL_X
                ),
                FixStatus.SYSTEM_STILL_VULNERABLE,
                [
                    UnfixedPackage(
                        pkg="pkg1",
                        unfixed_reason=messages.SECURITY_UA_SERVICE_NOT_ENABLED_SHORT.format(  # noqa
                            service="esm-infra"
                        ),
                    ),
                ],
            ),
        ),
    )
    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.apt.run_apt_command")
    @mock.patch("uaclient.system.should_reboot", return_value=False)
    @mock.patch("uaclient.util.prompt_choices", return_value="a")
    @mock.patch(M_PATH + "get_cloud_type", return_value=(None, None))
    @mock.patch("builtins.input", return_value="pro_token")
    @mock.patch(M_PATH + "attach_with_token")
    def test_execute_fix_plan_when_service_is_not_entitled(
        self,
        m_attach_with_token,
        _m_input,
        _m_get_cloud_type,
        _m_prompt,
        _m_should_reboot,
        _m_run_apt_command,
        _m_run_apt_update,
        fix_plan,
        expected_output,
        expected_fix_status,
        expected_unfixed_pkgs,
        service_status,
        fake_machine_token_file,
        capsys,
    ):
        def fake_attach(cfg, token, allow_enable):
            fake_machine_token_file.attached = True
            return 0

        m_attach_with_token.side_effect = fake_attach

        m_entitlement_obj = mock.MagicMock()
        m_entitlement_obj.user_facing_status.return_value = (
            service_status,
            "",
        )
        m_entitlement_obj.applicability_status.return_value = (
            ApplicabilityStatus.INAPPLICABLE,
            "",
        )
        m_entitlement_obj.name = "esm-infra"

        with mock.patch(
            "uaclient.cli.fix.entitlement_factory",
            return_value=m_entitlement_obj,
        ):
            with mock.patch("uaclient.util.sys") as m_sys:
                m_stdout = mock.MagicMock()
                type(m_sys).stdout = m_stdout
                type(m_stdout).encoding = mock.PropertyMock(
                    return_value="utf-8"
                )
                assert (
                    expected_fix_status,
                    expected_unfixed_pkgs,
                ) == execute_fix_plan(fix_plan, dry_run=False, cfg=None)

        out, _ = capsys.readouterr()
        assert expected_output in out

    @pytest.mark.parametrize(
        "fix_plan,prompt_value,expected_output,"
        "expected_fix_status,expected_unfixed_pkgs",
        (
            (  # installing package fixes that comes from
                # standard-updates and esm-infra when attaching
                # the system
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1"],
                    plan=[
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_INFRA_POCKET,
                                source_packages=["pkg1"],
                            ),
                            order=1,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg1"],
                                source_packages=["pkg1"],
                                pocket=ESM_INFRA_POCKET,
                            ),
                            order=2,
                        ),
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                "e",
                textwrap.dedent(
                    """\
                    1 affected source package is installed: pkg1
                    (1/1) pkg1:
                    A fix is available in Ubuntu Pro: ESM Infra.
                    """
                )
                + messages.SECURITY_SERVICE_DISABLED.format(
                    service="esm-infra"
                )
                + "\n"
                + colorize_commands([["pro enable esm-infra"]])
                + "\n"
                + colorize_commands(
                    [["apt update && apt install --only-upgrade" " -y pkg1"]]
                )
                + "\n\n"
                + "{check} USN-### is resolved.\n".format(
                    check=messages.OKGREEN_CHECK
                ),
                FixStatus.SYSTEM_NON_VULNERABLE,
                [],
            ),
            (
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1"],
                    plan=[
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_INFRA_POCKET,
                                source_packages=["pkg1"],
                            ),
                            order=1,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg1"],
                                source_packages=["pkg1"],
                                pocket=ESM_INFRA_POCKET,
                            ),
                            order=2,
                        ),
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                "c",
                textwrap.dedent(
                    """\
                    1 affected source package is installed: pkg1
                    (1/1) pkg1:
                    A fix is available in Ubuntu Pro: ESM Infra.
                    """
                )
                + messages.SECURITY_SERVICE_DISABLED.format(
                    service="esm-infra"
                )
                + "\n"
                + messages.SECURITY_UA_SERVICE_NOT_ENABLED.format(
                    service="esm-infra"
                )
                + "\n\n"
                + "1 package is still affected: pkg1"
                + "\n"
                + "{check} USN-### is not resolved.\n".format(
                    check=messages.FAIL_X
                ),
                FixStatus.SYSTEM_STILL_VULNERABLE,
                [
                    UnfixedPackage(
                        pkg="pkg1",
                        unfixed_reason=messages.SECURITY_UA_SERVICE_NOT_ENABLED_SHORT.format(  # noqa
                            service="esm-infra"
                        ),
                    )
                ],
            ),
        ),
    )
    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.apt.run_apt_command")
    @mock.patch(M_PATH + "enable_entitlement_by_name")
    @mock.patch("uaclient.system.should_reboot", return_value=False)
    @mock.patch("uaclient.util.prompt_choices")
    @mock.patch(M_PATH + "get_cloud_type", return_value=(None, None))
    @mock.patch("builtins.input", return_value="pro_token")
    def test_execute_fix_plan_when_service_requires_enable(
        self,
        _m_input,
        _m_get_cloud_type,
        m_prompt,
        _m_should_reboot,
        m_enable_ent,
        _m_run_apt_command,
        _m_run_apt_update,
        fix_plan,
        prompt_value,
        expected_output,
        expected_fix_status,
        expected_unfixed_pkgs,
        fake_machine_token_file,
        capsys,
    ):
        fake_machine_token_file.attached = True
        m_enable_ent.return_value = (True, None)
        m_prompt.return_value = prompt_value

        m_entitlement_obj = mock.MagicMock()
        m_entitlement_obj.user_facing_status.return_value = (
            UserFacingStatus.INACTIVE,
            "",
        )
        m_entitlement_obj.applicability_status.return_value = (
            ApplicabilityStatus.APPLICABLE,
            "",
        )
        m_entitlement_obj.name = "esm-infra"

        with mock.patch(
            "uaclient.cli.fix.entitlement_factory",
            return_value=m_entitlement_obj,
        ):
            with mock.patch("uaclient.util.sys") as m_sys:
                m_stdout = mock.MagicMock()
                type(m_sys).stdout = m_stdout
                type(m_stdout).encoding = mock.PropertyMock(
                    return_value="utf-8"
                )
                out, _ = capsys.readouterr()
                assert (
                    expected_fix_status,
                    expected_unfixed_pkgs,
                ) == execute_fix_plan(fix_plan, dry_run=False, cfg=None)

        out, _ = capsys.readouterr()
        assert expected_output in out

    @pytest.mark.parametrize(
        "fix_plan,prompt_value,expected_output,"
        "expected_fix_status,expected_unfixed_pkgs",
        (
            (
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1"],
                    plan=[
                        FixPlanAttachStep(
                            data=AttachData(
                                reason="test",
                                required_service=ESM_INFRA_POCKET,
                                source_packages=["pkg1"],
                            ),
                            order=1,
                        ),
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_INFRA_POCKET,
                                source_packages=["pkg1"],
                            ),
                            order=2,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg1"],
                                source_packages=["pkg1"],
                                pocket=ESM_INFRA_POCKET,
                            ),
                            order=3,
                        ),
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                "r",
                textwrap.dedent(
                    """\
                    1 affected source package is installed: pkg1
                    (1/1) pkg1:
                    A fix is available in Ubuntu Pro: ESM Infra.
                    """
                )
                + messages.SECURITY_UPDATE_NOT_INSTALLED_EXPIRED
                + "\n"
                + messages.PROMPT_EXPIRED_ENTER_TOKEN
                + "\n"
                + colorize_commands([["pro detach"]])
                + "\n"
                + colorize_commands([["pro attach pro_token"]])
                + "\n"
                + colorize_commands(
                    [["apt update && apt install --only-upgrade" " -y pkg1"]]
                )
                + "\n\n"
                + "{check} USN-### is resolved.\n".format(
                    check=messages.OKGREEN_CHECK
                ),
                FixStatus.SYSTEM_NON_VULNERABLE,
                [],
            ),
            (
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1"],
                    plan=[
                        FixPlanAttachStep(
                            data=AttachData(
                                reason="test",
                                required_service=ESM_INFRA_POCKET,
                                source_packages=["pkg1"],
                            ),
                            order=1,
                        ),
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_INFRA_POCKET,
                                source_packages=["pkg1"],
                            ),
                            order=2,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg1"],
                                source_packages=["pkg1"],
                                pocket=ESM_INFRA_POCKET,
                            ),
                            order=3,
                        ),
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                "c",
                textwrap.dedent(
                    """\
                    1 affected source package is installed: pkg1
                    (1/1) pkg1:
                    A fix is available in Ubuntu Pro: ESM Infra.
                    """
                )
                + messages.SECURITY_UPDATE_NOT_INSTALLED_EXPIRED
                + "\n\n"
                + "1 package is still affected: pkg1"
                + "\n"
                + "{check} USN-### is not resolved.\n".format(
                    check=messages.FAIL_X
                ),
                FixStatus.SYSTEM_STILL_VULNERABLE,
                [
                    UnfixedPackage(
                        pkg="pkg1",
                        unfixed_reason=messages.SECURITY_UA_SERVICE_WITH_EXPIRED_SUB.format(  # noqa
                            service="esm-infra"
                        ),
                    )
                ],
            ),
        ),
    )
    @mock.patch(M_PATH + "action_detach")
    @mock.patch(M_PATH + "attach_with_token")
    @mock.patch(M_PATH + "_check_subscription_is_expired")
    @mock.patch(M_PATH + "_handle_subscription_for_required_service")
    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.apt.run_apt_command")
    @mock.patch("uaclient.system.should_reboot", return_value=False)
    @mock.patch("uaclient.util.prompt_choices")
    @mock.patch(M_PATH + "get_cloud_type", return_value=(None, None))
    @mock.patch("builtins.input", return_value="pro_token")
    def test_execute_fix_plan_when_subscription_is_expired(
        self,
        _m_input,
        _m_get_cloud_type,
        m_prompt,
        _m_should_reboot,
        _m_run_apt_command,
        _m_run_apt_update,
        m_handle_required_service,
        m_check_subscription_expired,
        m_attach_with_token,
        _m_action_detach,
        fix_plan,
        prompt_value,
        expected_output,
        expected_fix_status,
        expected_unfixed_pkgs,
        fake_machine_token_file,
        capsys,
    ):
        fake_machine_token_file.attached = True
        m_handle_required_service.return_value = True
        m_check_subscription_expired.return_value = True
        m_prompt.return_value = prompt_value

        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            out, _ = capsys.readouterr()
            assert (
                expected_fix_status,
                expected_unfixed_pkgs,
            ) == execute_fix_plan(fix_plan, dry_run=False, cfg=None)

        out, _ = capsys.readouterr()
        assert expected_output in out

    @pytest.mark.parametrize(
        "fix_plan,check_notices,cloud_type,expected_output,"
        "expected_fix_status,expected_unfixed_pkgs",
        (
            (  # No affected_packages listed, but reboot required
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="not-affected",
                    expected_status=FixStatus.SYSTEM_NOT_AFFECTED.value.msg,
                    affected_packages=[],
                    plan=[
                        FixPlanNoOpStep(
                            data=NoOpData(
                                status=FixPlanNoOpStatus.NOT_AFFECTED.value
                            ),
                            order=1,
                        )
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                False,
                (None, None),
                textwrap.dedent(
                    """\
                   No affected source packages are installed.

                   {check} USN-### does not affect your system.
                   """.format(
                        check=messages.OKGREEN_CHECK
                    )
                ),
                FixStatus.SYSTEM_NOT_AFFECTED,
                [],
            ),
            (  # installing package fix and reboot required
                FixPlanResult(
                    title="USN-###",
                    description="test",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1"],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg1"],
                                source_packages=["pkg1"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=1,
                        )
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                True,
                (None, None),
                textwrap.dedent(
                    """\
                   1 affected source package is installed: pkg1
                   (1/1) pkg1:
                   A fix is available in Ubuntu standard updates.
                   """
                )
                + colorize_commands(
                    [["apt update && apt install --only-upgrade" " -y pkg1"]]
                )
                + "\n\n"
                + "A reboot is required to complete fix operation."
                + "\n"
                + "{check} USN-### is not resolved.\n".format(
                    check=messages.FAIL_X
                ),
                FixStatus.SYSTEM_VULNERABLE_UNTIL_REBOOT,
                [],
            ),
        ),
    )
    @mock.patch("uaclient.files.notices.NoticesManager.add")
    @mock.patch("uaclient.system.should_reboot", return_value=True)
    @mock.patch("uaclient.cli.fix.get_cloud_type")
    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.apt.run_apt_command")
    def test_execute_fix_plan_when_reboot_required_needed(
        self,
        _m_apt_run_cmd,
        _m_apt_run_update,
        m_get_cloud_type,
        _m_should_reboot,
        m_add_notice,
        fix_plan,
        check_notices,
        cloud_type,
        expected_output,
        expected_fix_status,
        expected_unfixed_pkgs,
        capsys,
        FakeConfig,
    ):
        m_get_cloud_type.return_value = cloud_type
        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            assert (
                expected_fix_status,
                expected_unfixed_pkgs,
            ) == execute_fix_plan(fix_plan, dry_run=False, cfg=FakeConfig())

        out, _ = capsys.readouterr()
        assert expected_output in out

        if check_notices:
            assert [
                mock.call(
                    Notice.ENABLE_REBOOT_REQUIRED,
                    messages.ENABLE_REBOOT_REQUIRED_TMPL.format(
                        operation="fix operation"
                    ),
                )
            ] == m_add_notice.call_args_list


class TestExecuteAptUpgradeStep:
    @mock.patch("uaclient.util.we_are_currently_root", return_value=False)
    def test_execute_apt_upgrade_step_return_early_if_non_root(
        self,
        _m_we_are_root,
        capsys,
    ):
        step = FixPlanAptUpgradeStep(
            data=AptUpgradeData(
                binary_packages=["pkg1"],
                source_packages=["pkg1"],
                pocket=STANDARD_UPDATES_POCKET,
            ),
            order=1,
        )
        context = FixContext(
            title="test",
            dry_run=False,
            affected_pkgs=[],
            cfg=None,
        )

        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            _execute_apt_upgrade_step(context, step)

        out, _ = capsys.readouterr()
        assert messages.SECURITY_APT_NON_ROOT in out


class TestExecuteAttachStep:
    def test_execute_attach_step_print_message_succeed_on_dry_run(
        self, capsys, FakeConfig
    ):
        step = FixPlanAttachStep(
            data=AttachData(
                reason="test",
                required_service=ESM_INFRA_POCKET,
                source_packages=["slsrc"],
            ),
            order=1,
        )
        context = FixContext(
            title="test",
            dry_run=True,
            affected_pkgs=[],
            cfg=FakeConfig(),
        )

        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            _execute_attach_step(context, step)

        out, _ = capsys.readouterr()
        assert messages.SECURITY_DRY_RUN_UA_NOT_ATTACHED in out
        assert context.fix_status == FixStatus.SYSTEM_NON_VULNERABLE


class TestExecuteEnableStep:
    def test_execute_enable_step_check_service_on_dry_run(
        self, capsys, FakeConfig
    ):
        step = FixPlanEnableStep(
            data=EnableData(
                service=ESM_INFRA_POCKET,
                source_packages=["slsrc"],
            ),
            order=1,
        )
        context = FixContext(
            title="test",
            dry_run=True,
            affected_pkgs=[],
            cfg=FakeConfig(),
        )

        m_entitlement_obj = mock.MagicMock()
        m_entitlement_obj.user_facing_status.return_value = (
            UserFacingStatus.INACTIVE,
            "",
        )
        m_entitlement_obj.applicability_status.return_value = (
            ApplicabilityStatus.APPLICABLE,
            "",
        )
        m_entitlement_obj.name = "esm-infra"

        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            with mock.patch(
                "uaclient.cli.fix.entitlement_factory",
                return_value=m_entitlement_obj,
            ):
                _execute_enable_step(context, step)

        out, _ = capsys.readouterr()
        assert (
            messages.SECURITY_DRY_RUN_UA_SERVICE_NOT_ENABLED.format(
                service="esm-infra"
            )
            in out
        )
        assert context.fix_status == FixStatus.SYSTEM_NON_VULNERABLE


class TestHandleSubscriptionForRequiredService:
    @pytest.mark.parametrize(
        "dry_run",
        ((True), (False)),
    )
    def test_handle_subscription_when_service_enabled(self, dry_run, capsys):
        m_entitlement_obj = mock.MagicMock()
        m_entitlement_obj.user_facing_status.return_value = (
            UserFacingStatus.ACTIVE,
            "",
        )
        m_entitlement_obj.name = "esm-infra"

        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            with mock.patch(
                "uaclient.cli.fix.entitlement_factory",
                return_value=m_entitlement_obj,
            ):
                assert _handle_subscription_for_required_service(
                    service="esm-infra",
                    cfg=None,
                    dry_run=dry_run,
                )
        out, _ = capsys.readouterr()
        assert "" == out


class TestPerformMagicAttach:
    @mock.patch(M_PATH + "_initiate")
    @mock.patch(M_PATH + "_wait")
    @mock.patch(M_PATH + "_revoke")
    def test_magic_attach_revoke_token_if_wait_fails(
        self,
        m_initiate,
        m_wait,
        m_revoke,
    ):
        m_initiate.return_value = mock.MagicMock(
            token="token", user_code="user_code"
        )
        m_wait.side_effect = exceptions.MagicAttachTokenError()

        with pytest.raises(exceptions.MagicAttachTokenError):
            _perform_magic_attach(cfg=None)

        assert 1 == m_initiate.call_count
        assert 1 == m_wait.call_count
        assert 1 == m_revoke.call_count


class TestFixUSN:
    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.apt.run_apt_command")
    @mock.patch(M_PATH + "_prompt_for_attach", return_value=False)
    @mock.patch(M_PATH + "usn_plan")
    def test_fix_usn_with_related_usns(
        self,
        m_usn_plan,
        _m_prompt_for_attach,
        _m_run_apt_command,
        _m_run_apt_update,
        capsys,
        FakeConfig,
    ):
        fix_plan = USNSFixPlanResult(
            usns_data=USNFixPlanResult(
                expected_status=FixStatus.SYSTEM_NON_VULNERABLE,
                usns=[
                    FixPlanUSNResult(
                        target_usn_plan=FixPlanResult(
                            title="USN-1235-1",
                            description="test",
                            current_status="still-affected",
                            expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,  # noqa
                            affected_packages=["pkg1"],
                            plan=[
                                FixPlanAptUpgradeStep(
                                    data=AptUpgradeData(
                                        binary_packages=["pkg1"],
                                        source_packages=["pkg1"],
                                        pocket=STANDARD_UPDATES_POCKET,
                                    ),
                                    order=1,
                                )
                            ],
                            warnings=[],
                            error=None,
                            additional_data=USNAdditionalData(
                                associated_cves=[],
                                associated_launchpad_bugs=[],
                            ),
                        ),
                        related_usns_plan=[
                            FixPlanResult(
                                title="USN-4561-1",
                                description="test",
                                current_status="still-affected",
                                expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,  # noqa
                                affected_packages=["pkg2"],
                                plan=[
                                    FixPlanAttachStep(
                                        data=AttachData(
                                            reason="test",
                                            required_service="esm-infra",
                                            source_packages=["pkg2"],
                                        ),
                                        order=1,
                                    ),
                                    FixPlanEnableStep(
                                        data=EnableData(
                                            service="esm-infra",
                                            source_packages=["pkg2"],
                                        ),
                                        order=2,
                                    ),
                                    FixPlanAptUpgradeStep(
                                        data=AptUpgradeData(
                                            binary_packages=["pkg2"],
                                            source_packages=["pkg2"],
                                            pocket=ESM_INFRA_POCKET,
                                        ),
                                        order=3,
                                    ),
                                ],
                                warnings=[],
                                error=None,
                                additional_data=USNAdditionalData(
                                    associated_cves=[],
                                    associated_launchpad_bugs=[],
                                ),
                            ),
                            FixPlanResult(
                                title="USN-7891-1",
                                description="test",
                                current_status="still-affected",
                                expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,  # noqa
                                affected_packages=["pkg3", "pkg4"],
                                plan=[
                                    FixPlanAttachStep(
                                        data=AttachData(
                                            reason="test",
                                            required_service=ESM_APPS_POCKET,
                                            source_packages=["pkg3", "pkg4"],
                                        ),
                                        order=1,
                                    ),
                                    FixPlanEnableStep(
                                        data=EnableData(
                                            service="esm-apps",
                                            source_packages=["pkg3", "pkg4"],
                                        ),
                                        order=2,
                                    ),
                                    FixPlanAptUpgradeStep(
                                        data=AptUpgradeData(
                                            binary_packages=["pkg3", "pkg4"],
                                            source_packages=["pkg3", "pkg4"],
                                            pocket=ESM_APPS_POCKET,
                                        ),
                                        order=3,
                                    ),
                                ],
                                warnings=[],
                                error=None,
                                additional_data=USNAdditionalData(
                                    associated_cves=[],
                                    associated_launchpad_bugs=[],
                                ),
                            ),
                            FixPlanResult(
                                title="USN-8221-1",
                                description="test",
                                current_status="still-affected",
                                expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,  # noqa
                                affected_packages=["pkg5"],
                                plan=[],
                                warnings=[
                                    FixPlanWarningSecurityIssueNotFixed(
                                        data=SecurityIssueNotFixedData(
                                            source_packages=["pkg5"],
                                            status="pending",
                                        ),
                                        order=1,
                                    ),
                                ],
                                error=None,
                                additional_data=USNAdditionalData(
                                    associated_cves=[],
                                    associated_launchpad_bugs=[],
                                ),
                            ),
                        ],
                    )
                ],
            )
        )
        m_usn_plan.return_value = fix_plan
        issue_id = "USN-1231-1"

        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            actual_ret = fix_usn(
                security_issue=issue_id,
                dry_run=False,
                no_related=False,
                cfg=FakeConfig(),
            )

        expected_msg = (
            "\n"
            + messages.SECURITY_FIXING_REQUESTED_USN.format(issue_id=issue_id)
            + "\n"
            + textwrap.dedent(
                """\
                1 affected source package is installed: pkg1
                (1/1) pkg1:
                A fix is available in Ubuntu standard updates.
                """
            )
            + colorize_commands(
                [["apt update && apt install --only-upgrade" " -y pkg1"]]
            )
            + "\n\n"
            + "{check} USN-1235-1 is resolved.\n".format(
                check=messages.OKGREEN_CHECK
            )
            + "\n"
            + textwrap.dedent(
                """\
                Found related USNs:
                - USN-4561-1
                - USN-7891-1
                - USN-8221-1
                """
            )
            + "\n"
            + textwrap.dedent(
                """\
                Fixing related USNs:
                - USN-4561-1
                1 affected source package is installed: pkg2
                (1/1) pkg2:
                A fix is available in Ubuntu Pro: ESM Infra.
                """
            )
            + "\n"
            + "1 package is still affected: pkg2"
            + "\n"
            + "{check} USN-4561-1 is not resolved.".format(
                check=messages.FAIL_X
            )
            + "\n\n"
            + textwrap.dedent(
                """\
                - USN-7891-1
                2 affected source packages are installed: pkg3, pkg4
                (1/2, 2/2) pkg3, pkg4:
                A fix is available in Ubuntu Pro: ESM Apps.
                """
            )
            + "\n"
            + "2 packages are still affected: pkg3, pkg4"
            + "\n"
            + "{check} USN-7891-1 is not resolved.".format(
                check=messages.FAIL_X
            )
            + "\n\n"
            + textwrap.dedent(
                """\
                - USN-8221-1
                1 affected source package is installed: pkg5
                (1/1) pkg5:
                A fix is coming soon. Try again tomorrow.
                """
            )
            + "\n"
            + "1 package is still affected: pkg5"
            + "\n"
            + "{check} USN-8221-1 is not resolved.".format(
                check=messages.FAIL_X
            )
            + "\n\n"
            + "Summary:"
            + "\n"
            + "{check} USN-1231-1 [requested] is resolved.".format(
                check=messages.OKGREEN_CHECK
            )
            + "\n"
            + "{check} USN-4561-1 [related] is not resolved.".format(
                check=messages.FAIL_X
            )
            + "\n"
            + "  - pkg2: esm-infra is required for upgrade."
            + "\n"
            + "{check} USN-7891-1 [related] is not resolved.".format(
                check=messages.FAIL_X
            )
            + "\n"
            + "  - pkg3: esm-apps is required for upgrade."
            + "\n"
            + "  - pkg4: esm-apps is required for upgrade."
            + "\n"
            + "{check} USN-8221-1 [related] is not resolved.".format(
                check=messages.FAIL_X
            )
            + "\n"
            + "  - pkg5: A fix is coming soon. Try again tomorrow."
            + "\n\n"
            + messages.SECURITY_RELATED_USN_ERROR.format(issue_id="USN-1231-1")
        )
        out, _ = capsys.readouterr()
        assert expected_msg in out
        assert FixStatus.SYSTEM_NON_VULNERABLE == actual_ret

    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.apt.run_apt_command")
    @mock.patch(M_PATH + "usn_plan")
    def test_fix_usn_when_no_related_value_is_true(
        self,
        m_usn_plan,
        _m_run_apt_command,
        _m_run_apt_update,
        capsys,
        FakeConfig,
    ):
        issue_id = "USN-1235-1"
        fix_plan = USNSFixPlanResult(
            usns_data=USNFixPlanResult(
                expected_status=FixStatus.SYSTEM_NON_VULNERABLE,
                usns=[
                    FixPlanUSNResult(
                        target_usn_plan=FixPlanResult(
                            title=issue_id,
                            description="test",
                            current_status="still-affected",
                            expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,  # noqa
                            affected_packages=["pkg1"],
                            plan=[
                                FixPlanAptUpgradeStep(
                                    data=AptUpgradeData(
                                        binary_packages=["pkg1"],
                                        source_packages=["pkg1"],
                                        pocket=STANDARD_UPDATES_POCKET,
                                    ),
                                    order=1,
                                )
                            ],
                            warnings=[],
                            error=None,
                            additional_data=USNAdditionalData(
                                associated_cves=[],
                                associated_launchpad_bugs=[],
                            ),
                        ),
                        related_usns_plan=[
                            FixPlanResult(
                                title="USN-4561-1",
                                description="test",
                                current_status="still-affected",
                                expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,  # noqa
                                affected_packages=["pkg2"],
                                plan=[
                                    FixPlanAttachStep(
                                        data=AttachData(
                                            reason="test",
                                            required_service="esm-infra",
                                            source_packages=["pkg2"],
                                        ),
                                        order=1,
                                    ),
                                    FixPlanEnableStep(
                                        data=EnableData(
                                            service="esm-infra",
                                            source_packages=["pkg2"],
                                        ),
                                        order=2,
                                    ),
                                    FixPlanAptUpgradeStep(
                                        data=AptUpgradeData(
                                            binary_packages=["pkg2"],
                                            source_packages=["pkg2"],
                                            pocket=ESM_INFRA_POCKET,
                                        ),
                                        order=3,
                                    ),
                                ],
                                warnings=[],
                                error=None,
                                additional_data=USNAdditionalData(
                                    associated_cves=[],
                                    associated_launchpad_bugs=[],
                                ),
                            ),
                            FixPlanResult(
                                title="USN-7891-1",
                                description="test",
                                current_status="still-affected",
                                expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,  # noqa
                                affected_packages=["pkg3", "pkg4"],
                                plan=[
                                    FixPlanAttachStep(
                                        data=AttachData(
                                            reason="test",
                                            required_service=ESM_APPS_POCKET,
                                            source_packages=["pkg3", "pkg4"],
                                        ),
                                        order=1,
                                    ),
                                    FixPlanEnableStep(
                                        data=EnableData(
                                            service="esm-apps",
                                            source_packages=["pkg3", "pkg4"],
                                        ),
                                        order=2,
                                    ),
                                    FixPlanAptUpgradeStep(
                                        data=AptUpgradeData(
                                            binary_packages=["pkg3", "pkg4"],
                                            source_packages=["pkg3", "pkg4"],
                                            pocket=ESM_APPS_POCKET,
                                        ),
                                        order=3,
                                    ),
                                ],
                                warnings=[],
                                error=None,
                                additional_data=USNAdditionalData(
                                    associated_cves=[],
                                    associated_launchpad_bugs=[],
                                ),
                            ),
                        ],
                    )
                ],
            )
        )
        m_usn_plan.return_value = fix_plan

        with mock.patch("uaclient.util.sys") as m_sys:
            m_stdout = mock.MagicMock()
            type(m_sys).stdout = m_stdout
            type(m_stdout).encoding = mock.PropertyMock(return_value="utf-8")
            actual_ret = fix_usn(
                security_issue=issue_id,
                dry_run=False,
                no_related=True,
                cfg=FakeConfig(),
            )

        expected_msg = (
            "USN-1235-1: test"
            + "\n\n"
            + messages.SECURITY_FIXING_REQUESTED_USN.format(issue_id=issue_id)
            + "\n"
            + textwrap.dedent(
                """\
                1 affected source package is installed: pkg1
                (1/1) pkg1:
                A fix is available in Ubuntu standard updates.
                """
            )
            + colorize_commands(
                [["apt update && apt install --only-upgrade" " -y pkg1"]]
            )
            + "\n\n"
            + "{check} USN-1235-1 is resolved.\n".format(
                check=messages.OKGREEN_CHECK
            )
        )

        out, _ = capsys.readouterr()
        assert expected_msg == out
        assert FixStatus.SYSTEM_NON_VULNERABLE == actual_ret



from uaclient.api.u.pro.security.fix._common.plan.v1 import (
    AptUpgradeData,
    AttachData,
    EnableData,
    FixPlanError,
    FixPlanResult,
    FixPlanStep,
    FixPlanWarning,
    NoOpData,
    PackageCannotBeInstalledData,
    SecurityIssueNotFixedData,
)

__all__ = [
    "AptUpgradeData",
    "AttachData",
    "EnableData",
    "FixPlanError",
    "FixPlanResult",
    "FixPlanStep",
    "FixPlanWarning",
    "NoOpData",
    "PackageCannotBeInstalledData",
    "SecurityIssueNotFixedData",
]



import logging
from typing import Iterable, List, Optional, Tuple, Type

from uaclient import entitlements, lock, messages, status, util
from uaclient.api import AbstractProgress, ProgressWrapper, exceptions
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo, ErrorWarningObject
from uaclient.api.u.pro.status.enabled_services.v1 import _enabled_services
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.config import UAConfig
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    Field,
    StringDataValue,
    data_list,
)

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


class EnableOptions(DataObject):
    fields = [
        Field("service", StringDataValue, doc="Pro service to be enabled"),
        Field(
            "variant",
            StringDataValue,
            False,
            doc="Optional variant of the Pro service to be enabled.",
        ),
        Field(
            "access_only",
            BoolDataValue,
            False,
            doc=(
                "If true and the target service supports it, only enable"
                " access to the service (default: false)"
            ),
        ),
    ]

    def __init__(
        self,
        *,
        service: str,
        variant: Optional[str] = None,
        access_only: bool = False
    ):
        self.service = service
        self.variant = variant
        self.access_only = access_only


class EnableResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "enabled",
            data_list(StringDataValue),
            doc="List of services that were enabled.",
        ),
        Field(
            "disabled",
            data_list(StringDataValue),
            doc="List of services that were disabled.",
        ),
        Field(
            "reboot_required",
            BoolDataValue,
            doc=(
                "True if one of the services that was enabled requires a"
                " reboot."
            ),
        ),
        Field(
            "messages",
            data_list(StringDataValue),
            doc=(
                "List of information message strings about the service that"
                " was just enabled. Possibly translated."
            ),
        ),
    ]

    def __init__(
        self,
        *,
        enabled: List[str],
        disabled: List[str],
        reboot_required: bool,
        messages: List[str]
    ):
        self.enabled = enabled
        self.disabled = disabled
        self.reboot_required = reboot_required
        self.messages = messages


def _auto_select_variant(
    cfg: UAConfig,
    progress: ProgressWrapper,
    entitlement: entitlements.UAEntitlement,
    available_variants: Iterable[Type[entitlements.UAEntitlement]],
    access_only: bool,
) -> Tuple[entitlements.UAEntitlement, Optional[ErrorWarningObject]]:
    variant = None
    for v_cls in available_variants:
        v = v_cls(cfg=cfg, access_only=access_only)
        if (
            v.applicability_status()[0]
            == entitlements.ApplicabilityStatus.APPLICABLE
            and v.variant_auto_select()
        ):
            variant = v
            break
    if variant is None and entitlement.default_variant is not None:
        variant = entitlement.default_variant(cfg=cfg, access_only=access_only)
    if variant is not None:
        progress.emit(
            "message_operation",
            [
                (
                    util.prompt_for_confirmation,
                    {
                        "msg": messages.AUTO_SELECTING_VARIANT.format(
                            variant=messages.TxtColor.BOLD
                            + variant.variant_name
                            + messages.TxtColor.ENDC
                        ),
                    },
                )
            ],
        )
        warning = messages.AUTO_SELECTED_VARIANT_WARNING.format(
            variant_name=variant.variant_name
        )
        return variant, ErrorWarningObject(
            title=warning.msg,
            code=warning.name,
            meta={"variant_name": variant.variant_name},
        )
    else:
        return entitlement, None


def _enabled_services_names(cfg: UAConfig) -> List[str]:
    return [s.name for s in _enabled_services(cfg).enabled_services]


def enable(
    options: EnableOptions, progress_object: Optional[AbstractProgress] = None
) -> EnableResult:
    return _enable(options, UAConfig(), progress_object=progress_object)


def _enable(
    options: EnableOptions,
    cfg: UAConfig,
    progress_object: Optional[AbstractProgress] = None,
) -> EnableResult:
    """
    Enable a Pro service. This will automatically disable incompatible services
    and enable required services that that target service depends on.
    """
    progress = ProgressWrapper(progress_object)
    warnings = []

    if not util.we_are_currently_root():
        raise exceptions.NonRootUserError()

    if not _is_attached(cfg).is_attached:
        raise exceptions.UnattachedError()

    if options.service == "landscape":
        raise exceptions.NotSupported()

    enabled_services_before = _enabled_services_names(cfg)

    already_enabled = next(
        (
            s
            for s in _enabled_services(cfg).enabled_services
            if s.name == options.service
            and (
                not options.variant
                or (s.variant_enabled and s.variant_name == options.variant)
            )
        ),
        None,
    )
    if already_enabled:
        # nothing to do
        return EnableResult(
            enabled=[],
            disabled=[],
            reboot_required=False,
            messages=[],
        )

    entitlement = entitlements.entitlement_factory(
        cfg=cfg,
        name=options.service,
        variant=options.variant or "",
        access_only=options.access_only,
    )

    applicability, _ = entitlement.applicability_status()
    available_variants = entitlement.variants
    if all(
        [
            applicability == entitlements.ApplicabilityStatus.APPLICABLE,
            not entitlement.is_variant,
            available_variants,
        ]
    ):
        entitlement, auto_select_warning = _auto_select_variant(
            cfg=cfg,
            progress=progress,
            entitlement=entitlement,
            available_variants=available_variants.values(),
            access_only=options.access_only,
        )
        if auto_select_warning:
            warnings.append(auto_select_warning)

    progress.total_steps = entitlement.calculate_total_enable_steps()

    success = False
    fail_reason = None

    try:
        with lock.RetryLock(
            lock_holder="u.pro.services.enable.v1",
        ):
            success, fail_reason = entitlement.enable(progress)
    except Exception as e:
        lock.clear_lock_file_if_present()
        raise e

    if not success:
        if fail_reason is not None and fail_reason.message is not None:
            reason = fail_reason.message
        else:
            reason = messages.GENERIC_UNKNOWN_ISSUE
        raise exceptions.EntitlementNotEnabledError(
            service=options.service, reason=reason
        )

    enabled_services_after = _enabled_services_names(cfg)

    post_enable_messages = [
        msg
        for msg in (entitlement.messaging.get("post_enable", []) or [])
        if isinstance(msg, str)
    ]

    status.status(cfg=cfg)  # Update the status cache
    progress.finish()

    result = EnableResult(
        enabled=sorted(
            list(
                set(enabled_services_after).difference(
                    set(enabled_services_before)
                )
            )
        ),
        disabled=sorted(
            list(
                set(enabled_services_before).difference(
                    set(enabled_services_after)
                )
            )
        ),
        reboot_required=entitlement._check_for_reboot(),
        messages=post_enable_messages,
    )
    result.warnings = warnings
    return result


endpoint = APIEndpoint(
    version="v1",
    name="EnableService",
    fn=_enable,
    options_cls=EnableOptions,
    supports_progress=True,
)

_doc = {
    "introduced_in": "32",
    "requires_network": True,
    "example_python": """
from uaclient.api.u.pro.services.enable.v1 import enable, EnableOptions
result = enable(EnableOptions(service="usg"))
""",
    "result_class": EnableResult,
    "exceptions": [
        (exceptions.NonRootUserError, "When called as non-root user"),
        (
            exceptions.UnattachedError,
            (
                "When called on a machine that is not attached to a Pro"
                " subscription"
            ),
        ),
        (
            exceptions.NotSupported,
            (
                "When called for a service that doesn't support being enabled"
                " via API (currently only Landscape)"
            ),
        ),
        (
            exceptions.EntitlementNotFoundError,
            (
                "When the service argument is not a valid Pro service name or"
                " if the variant is not a valid variant of the target service"
            ),
        ),
        (
            exceptions.LockHeldError,
            "When another Ubuntu Pro related operation is in progress",
        ),
        (
            exceptions.EntitlementNotEnabledError,
            "When the service fails to enable",
        ),
    ],
    "example_cli": "pro api u.pro.services.enable.v1 --args service=usg",
    "example_json": """
{
    "disabled": [],
    "enabled": [
        "usg"
    ],
    "messages": [],
    "reboot_required": false
}
""",
}



import logging
import os
import shutil

from uaclient import exceptions, util

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


def export_gpg_key(source_keyfile: str, destination_keyfile: str) -> None:
    """Copy a specific key from source_keyring_dir into destination_keyfile

    :param source_keyfile: Path of source keyring file to export.
    :param destination_keyfile: The filename created with the single exported
        key.

    :raise UbuntuProError: Any GPG errors or if specific key does not exist in
        the source_keyring_file.
    """
    LOG.debug("Exporting GPG key %s", source_keyfile)
    if not os.path.exists(source_keyfile):
        raise exceptions.GPGKeyNotFound(keyfile=source_keyfile)
    shutil.copy(source_keyfile, destination_keyfile)
    os.chmod(destination_keyfile, 0o644)



import datetime

import mock
import pytest

from uaclient import exceptions
from uaclient.update_contract_info import validate_release_series

M_PATH = "uaclient.update_contract_info."


class TestValidateReleaseSeries:
    @pytest.mark.parametrize(
        "only_series,eol,is_valid",
        (
            ("bionic", datetime.date(2023, 5, 31), False),
            ("jammy", datetime.date(2027, 6, 1), True),
            ("noble", datetime.date(2029, 5, 31), True),
        ),
    )
    @mock.patch(
        "uaclient.system.get_release_info",
        return_value=mock.MagicMock(series="jammy"),
    )
    @mock.patch("uaclient.system.get_distro_info")
    @mock.patch(M_PATH + "detach")
    @mock.patch(
        M_PATH + "_is_attached",
        return_value=mock.MagicMock(is_attached=True),
    )
    def test_validate_release_series(
        self,
        m_is_attached,
        m_detach,
        m_get_distro_info,
        m_get_release_info,
        only_series,
        eol,
        is_valid,
        FakeConfig,
        fake_machine_token_file,
    ):
        fake_machine_token_file._entitlements = {
            "support": {
                "entitlement": {"affordances": {"onlySeries": only_series}}
            }
        }
        m_get_distro_info.side_effect = [
            mock.MagicMock(
                series=only_series,
                eol=eol,
            ),
            mock.MagicMock(
                series="jammy",
                eol=datetime.date(2027, 6, 1),
            ),
        ]
        validate_release_series(cfg=FakeConfig(), only_series=only_series)
        if is_valid:
            assert m_get_distro_info.call_count == 2
            assert m_detach.call_count == 0
        else:
            assert m_get_distro_info.call_count == 2
            assert m_detach.call_count == 1

    @mock.patch(
        "uaclient.system.get_release_info",
        return_value=mock.MagicMock(series="jammy"),
    )
    @mock.patch("uaclient.system.get_distro_info")
    @mock.patch(M_PATH + "detach")
    @mock.patch(
        M_PATH + "_is_attached",
        return_value=mock.MagicMock(is_attached=True),
    )
    def test_validate_release_series_not_found(
        self,
        m_is_attached,
        m_detach,
        m_get_distro_info,
        m_get_release_info,
        FakeConfig,
    ):
        m_get_distro_info.side_effect = (
            exceptions.MissingSeriesInDistroInfoFile(series="plucky")
        )
        validate_release_series(cfg=FakeConfig(), only_series="plucky")
        assert m_get_distro_info.call_count == 1
        assert m_detach.call_count == 0



import mock
import pytest

from lib.migrate_user_config import (
    create_new_uaclient_conffile,
    create_new_user_config_file,
    load_pre_upgrade_conf,
)
from uaclient import messages
from uaclient.testing.fakes import FakeFile
from uaclient.yaml import safe_load


class TestLoadPreUpgradeConf:
    @pytest.mark.parametrize(
        ["contents"],
        [
            (
                """\
contracts_url: urlhere
log_level: debug
features: {allow_beta: true}
""",
            )
        ],
    )
    @mock.patch("builtins.open")
    def test_success(self, m_open, contents, capsys):
        m_open.return_value = FakeFile(contents)
        assert safe_load(contents) == load_pre_upgrade_conf()
        assert ("", "") == capsys.readouterr()

    @pytest.mark.parametrize(
        ["contents"],
        [
            (
                """\
contracts_url: {invalid
""",
            )
        ],
    )
    @mock.patch("builtins.open")
    def test_invalid_yaml(self, m_open, contents, capsys):
        m_open.return_value = FakeFile(contents)
        assert load_pre_upgrade_conf() is None
        assert (
            "",
            messages.USER_CONFIG_MIGRATION_WARNING_UACLIENT_CONF_LOAD + "\n",
        ) == capsys.readouterr()

    @mock.patch("builtins.open")
    def test_file_access_error(self, m_open, capsys):
        m_open.side_effect = Exception()
        assert load_pre_upgrade_conf() is None
        assert (
            "",
            messages.USER_CONFIG_MIGRATION_WARNING_UACLIENT_CONF_LOAD + "\n",
        ) == capsys.readouterr()


class TestCreateNewUserConfigFile:
    @pytest.mark.parametrize(
        ["old_conf", "expected_msg"],
        [
            ({}, ""),
            ({"log_level": "warning"}, ""),
            ({"ua_config": "invalid"}, ""),
            ({"ua_config": {"apt_news": True}}, ""),
            (
                {"ua_config": {"apt_news": False}},
                messages.USER_CONFIG_MIGRATION_WARNING_NEW_USER_CONFIG_WRITE
                + "\n           pro config set apt_news=False\n",
            ),
            (
                {"ua_config": {"apt_news": False, "http_proxy": "custom"}},
                messages.USER_CONFIG_MIGRATION_WARNING_NEW_USER_CONFIG_WRITE
                + "\n"
                + "           pro config set apt_news=False\n"
                + "           pro config set http_proxy=custom\n",
            ),
        ],
    )
    @mock.patch("builtins.open")
    def test_file_access_error(self, m_open, old_conf, expected_msg, capsys):
        m_open.side_effect = Exception()
        create_new_user_config_file(old_conf)
        assert ("", expected_msg) == capsys.readouterr()

    @pytest.mark.parametrize(
        ["old_conf", "new_user_config"],
        [
            ({}, {}),
            ({"log_level": "warning"}, {}),
            ({"ua_config": "invalid"}, {}),
            ({"ua_config": {"apt_news": True}}, {}),
            ({"ua_config": {"apt_news": False}}, {"apt_news": False}),
            (
                {"ua_config": {"apt_news": False, "http_proxy": "custom"}},
                {"http_proxy": "custom", "apt_news": False},
            ),
        ],
    )
    @mock.patch("json.dump")
    @mock.patch("builtins.open")
    def test_success(
        self, m_open, m_json_dump, old_conf, new_user_config, capsys
    ):
        create_new_user_config_file(old_conf)
        assert [
            mock.call(new_user_config, mock.ANY)
        ] == m_json_dump.call_args_list
        assert ("", "") == capsys.readouterr()


class TestCreateNewUaclientConffile:
    @pytest.mark.parametrize(
        ["old_conf", "expected_msg"],
        [
            (
                {},
                messages.USER_CONFIG_MIGRATION_MIGRATING
                + "\n"
                + messages.USER_CONFIG_MIGRATION_WARNING_NEW_UACLIENT_CONF_WRITE  # noqa: E501
                + "\n"
                + "           contract_url: 'https://contracts.canonical.com'\n"  # noqa: E501
                + "           log_level: 'debug'\n",
            ),
            (
                {"log_level": "debug"},
                messages.USER_CONFIG_MIGRATION_MIGRATING
                + "\n"
                + messages.USER_CONFIG_MIGRATION_WARNING_NEW_UACLIENT_CONF_WRITE  # noqa: E501
                + "\n"
                + "           contract_url: 'https://contracts.canonical.com'\n"  # noqa: E501
                + "           log_level: 'debug'\n",
            ),
            (
                {"log_level": "warning"},
                messages.USER_CONFIG_MIGRATION_MIGRATING
                + "\n"
                + messages.USER_CONFIG_MIGRATION_WARNING_NEW_UACLIENT_CONF_WRITE  # noqa: E501
                + "\n"
                + "           contract_url: 'https://contracts.canonical.com'\n"  # noqa: E501
                + "           log_level: 'warning'\n",
            ),
            (
                {"features": {}},
                messages.USER_CONFIG_MIGRATION_MIGRATING
                + "\n"
                + messages.USER_CONFIG_MIGRATION_WARNING_NEW_UACLIENT_CONF_WRITE  # noqa: E501
                + "\n"
                + "           contract_url: 'https://contracts.canonical.com'\n"  # noqa: E501
                + "           features: {}\n"
                + "           log_level: 'debug'\n",
            ),
        ],
    )
    @mock.patch("os.rename")
    @mock.patch("builtins.open")
    def test_file_access_error(
        self, m_open, _m_rename, old_conf, expected_msg, capsys
    ):
        m_open.side_effect = Exception()
        create_new_uaclient_conffile(old_conf)
        assert ("", expected_msg) == capsys.readouterr()

    @pytest.mark.parametrize(
        ["old_conf", "new_uaclient_conf"],
        [
            (
                {},
                {
                    "contract_url": "https://contracts.canonical.com",
                    "log_level": "debug",
                },
            ),
            (
                {"log_level": "warning"},
                {
                    "contract_url": "https://contracts.canonical.com",
                    "log_level": "warning",
                },
            ),
            (
                {"features": {"allow_beta": True}},
                {
                    "contract_url": "https://contracts.canonical.com",
                    "log_level": "debug",
                    "features": {"allow_beta": True},
                },
            ),
            (
                {
                    "contract_url": "https://contracts.canonical.com",
                    "log_level": "debug",
                    "log_file": "/var/log/ubuntu-advantage.log",
                    "daemon_log_file": "/var/log/ubuntu-advantage-daemon.log",
                    "timer_log_file": "/var/log/ubuntu-advantage-timer.log",
                    "data_dir": "/var/lib/ubuntu-advantage",
                    "ua_config": {"apt_news": False},
                },
                {
                    "contract_url": "https://contracts.canonical.com",
                    "log_level": "debug",
                },
            ),
            (
                {
                    "contract_url": "https://contracts.canonical.com",
                    "log_level": "debug",
                    "log_file": "/var/log/ubuntu-advantage.log",
                    "data_dir": "/var/lib/custom",
                },
                {
                    "contract_url": "https://contracts.canonical.com",
                    "log_level": "debug",
                    "data_dir": "/var/lib/custom",
                },
            ),
        ],
    )
    @mock.patch("os.rename")
    @mock.patch("lib.migrate_user_config.safe_dump")
    @mock.patch("builtins.open")
    def test_success(
        self,
        m_open,
        m_yaml_dump,
        _m_rename,
        old_conf,
        new_uaclient_conf,
        capsys,
    ):
        create_new_uaclient_conffile(old_conf)
        assert [
            mock.call(new_uaclient_conf, default_flow_style=False)
        ] == m_yaml_dump.call_args_list
        assert (
            "",
            messages.USER_CONFIG_MIGRATION_MIGRATING + "\n",
        ) == capsys.readouterr()



from uaclient import messages
from uaclient.api.errors import APIError
from uaclient.exceptions import (
    AlreadyAttachedError,
    ConnectivityError,
    ContractAPIError,
    DepedentOptionError,
    EntitlementNotDisabledError,
    EntitlementNotEnabledError,
    EntitlementNotFoundError,
    EntitlementsNotEnabledError,
    FeatureNotSupportedOldTokenError,
    IncompatibleServiceStopsEnable,
    InvalidOptionCombination,
    InvalidProImage,
    LockHeldError,
    MagicAttachTokenAlreadyActivated,
    MagicAttachTokenError,
    MagicAttachUnavailable,
    NonAutoAttachImageError,
    NonRootUserError,
    RequiredServiceStopsEnable,
    UbuntuProError,
    UnattachedError,
    UnsupportedManifestFile,
    UrlError,
    UserFacingError,
)

__all__ = [
    "AlreadyAttachedError",
    "ConnectivityError",
    "ContractAPIError",
    "EntitlementNotFoundError",
    "InvalidOptionCombination",
    "FeatureNotSupportedOldTokenError",
    "InvalidProImage",
    "LockHeldError",
    "MagicAttachTokenAlreadyActivated",
    "MagicAttachTokenError",
    "MagicAttachUnavailable",
    "NonAutoAttachImageError",
    "NonRootUserError",
    "UbuntuProError",
    "UnattachedError",
    "UrlError",
    "UserFacingError",
    "EntitlementsNotEnabledError",
    "EntitlementNotEnabledError",
    "EntitlementNotDisabledError",
    "IncompatibleServiceStopsEnable",
    "RequiredServiceStopsEnable",
    "UnsupportedManifestFile",
    "DepedentOptionError",
]


class AutoAttachDisabledError(UbuntuProError):
    _msg = messages.E_AUTO_ATTACH_DISABLED_ERROR


class UnattendedUpgradesError(APIError):
    _formatted_msg = messages.E_UNATTENDED_UPGRADES_ERROR


class NotSupported(UbuntuProError):
    _msg = messages.E_NOT_SUPPORTED



import logging

from behave import then

from features.steps.files import when_i_create_file_with_content
from features.steps.shell import when_i_run_command, when_i_run_shell_command


@then("`{file_name}` is not present in any docker image layer")
def file_is_not_present_in_any_docker_image_layer(context, file_name):
    when_i_run_command(
        context,
        "find /var/lib/docker/overlay2 -name {}".format(file_name),
        "with sudo",
    )
    results = context.process.stdout.strip()
    if results:
        raise AssertionError(
            'found "{}"'.format(", ".join(results.split("\n")))
        )


# This defines "not significantly larger" as "less than 2MB larger"
@then(
    "docker image `{name}` is not significantly larger than `ubuntu:{series}` with `{package}` installed"  # noqa: E501
)
def docker_image_is_not_larger(context, name, series, package):
    base_image_name = "ubuntu:{}".format(series)
    base_upgraded_image_name = "{}-with-test-package".format(series)

    # We need to compare against the base image after apt upgrade
    # and package install
    dockerfile = """\
    FROM {}
    RUN apt-get update \\
      && apt-get install -y {} \\
      && rm -rf /var/lib/apt/lists/*
    """.format(
        base_image_name, package
    )
    context.text = dockerfile
    when_i_create_file_with_content(context, "Dockerfile.base")
    when_i_run_command(
        context,
        "docker build . -f Dockerfile.base -t {}".format(
            base_upgraded_image_name
        ),
        "with sudo",
    )

    # find image sizes
    when_i_run_shell_command(
        context, "docker inspect {} | jq .[0].Size".format(name), "with sudo"
    )
    custom_image_size = int(context.process.stdout.strip())
    when_i_run_shell_command(
        context,
        "docker inspect {} | jq .[0].Size".format(base_upgraded_image_name),
        "with sudo",
    )
    base_image_size = int(context.process.stdout.strip())

    # Get pro test deb size
    when_i_run_command(context, "du ubuntu-advantage-tools.deb", "with sudo")
    # Example out: "1234\tubuntu-advantage-tools.deb"
    ua_test_deb_size = (
        int(context.process.stdout.strip().split("\t")[0]) * 1024
    )  # KB -> B

    # Give us some space for bloat we don't control: 2MB -> B
    extra_space = 2 * 1024 * 1024

    if custom_image_size > (base_image_size + ua_test_deb_size + extra_space):
        raise AssertionError(
            "Custom image size ({}) is over 2MB greater than the base image"
            " size ({}) + pro test deb size ({})".format(
                custom_image_size, base_image_size, ua_test_deb_size
            )
        )
    logging.debug(
        "custom image size ({})\n"
        "base image size ({})\n"
        "pro test deb size ({})".format(
            custom_image_size, base_image_size, ua_test_deb_size
        )
    )



import textwrap

import mock
import pytest

from uaclient.cli.formatter import Block, ContentAlignment
from uaclient.cli.formatter import ProOutputFormatterConfig as POFC
from uaclient.cli.formatter import Table, create_link, real_len, wrap_text

M_PATH = "uaclient.cli.formatter."


class TestProFormatterConfig:
    @pytest.mark.parametrize(
        "system_encoding,expected_use_utf8",
        ((None, False), ("iso-8859-13", False), ("utf-8", True)),
    )
    def test_use_utf8(self, system_encoding, expected_use_utf8, FakeConfig):
        cfg = FakeConfig()

        with mock.patch(M_PATH + "sys.stdout") as m_stdout:
            m_stdout.encoding = system_encoding
            POFC.init(cfg)

        assert POFC.use_utf8 is expected_use_utf8

    @pytest.mark.parametrize("is_tty", (True, False))
    @pytest.mark.parametrize("env_no_color", ("True", None))
    @mock.patch(M_PATH + "sys.stdout.isatty")
    @mock.patch(M_PATH + "os.getenv")
    def test_color_config(
        self,
        m_getenv,
        m_is_tty,
        is_tty,
        env_no_color,
        FakeConfig,
    ):
        cfg = FakeConfig()

        m_is_tty.return_value = is_tty

        m_getenv.return_value = env_no_color

        POFC.init(cfg)

        expected_result = True
        if any((not is_tty, env_no_color)):
            expected_result = False

        assert POFC.use_color is expected_result

        POFC.disable_color()
        assert POFC.use_color is False


class TestCreateLink:
    def test_creates_link(self):
        assert (
            "\x1b]8;;http://example.test\x1b\\some link\x1b]8;;\x1b\\"
            == create_link(text="some link", url="http://example.test")
        )


class TestRealLen:
    @pytest.mark.parametrize(
        "input_string,expected_length",
        (
            ("", 0),
            ("input text", 10),
            ("\033[1mbold text\033[0m", 9),
            ("\033[91mred text\033[0m", 8),
            ("\033[1m\033[91mbold and red text\033[0m", 17),
            ("some \033[1mbold\033[0m and some \033[91mred\033[0m text", 27),
        ),
    )
    def test_length_ignores_color(self, input_string, expected_length):
        assert expected_length == real_len(input_string)

    def test_length_ignores_links(self):
        str_with_link = (
            "There is an "
            "\x1b]8;;https://example.com\x1b\\example link here\x1b]8;;\x1b\\"
            " to be clicked"
        )
        assert 43 == real_len(str_with_link)


class TestWrapText:
    def test_single_line_wrapped(self):
        assert ["example"] == wrap_text("example", 20)

        long_string = (
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. "
            "Pellentesque diam nulla, efficitur a orci non, "
            "scelerisque lobortis felis."
        )
        assert [
            "Lorem ipsum dolor sit amet,",
            "consectetur adipiscing elit.",
            "Pellentesque diam nulla,",
            "efficitur a orci non,",
            "scelerisque lobortis felis.",
        ] == wrap_text(long_string, 30)

        colored_string = (
            "some \033[1mbold\033[0m and include "
            "the \033[91mred\033[0m text too"
        )
        assert len(colored_string) == 55
        assert [colored_string] == wrap_text(colored_string, 40)
        assert [
            "some \x1b[1mbold\x1b[0m and",
            "include the \x1b[91mred\x1b[0m text",
            "too",
        ] == wrap_text(colored_string, 20)

        string_with_link = (
            "There is an "
            "\x1b]8;;https://example.com\x1b\\example link here\x1b]8;;\x1b\\"
            " to be clicked"
        )
        assert len(string_with_link) == 76
        assert [string_with_link] == wrap_text(string_with_link, 45)
        assert [
            "There is an \x1b]8;;https://example.com\x1b\\example",
            "link here\x1b]8;;\x1b\\ to be",
            "clicked",
        ] == wrap_text(string_with_link, 20)


class TestTable:
    @pytest.mark.parametrize(
        "input_str,input_len,expected_value",
        (
            ("", 0, ""),
            ("", 5, "     "),
            ("test", 2, "test"),
            ("test", 4, "test"),
            ("test", 10, "test      "),
        ),
    )
    def test_ljust(self, input_str, input_len, expected_value):
        assert expected_value == Table.ljust(input_str, input_len)

    @pytest.mark.parametrize(
        "input_str,input_len,expected_value",
        (
            ("", 0, ""),
            ("", 5, "     "),
            ("test", 2, "test"),
            ("test", 4, "test"),
            ("test", 10, "      test"),
        ),
    )
    def test_rjust(self, input_str, input_len, expected_value):
        assert expected_value == Table.rjust(input_str, input_len)

    @pytest.mark.parametrize(
        "headers,rows,alignment,expected_msg",
        (
            (None, None, None, "Empty table not supported."),
            (
                None,
                [["a", "b"], [], ["c", "d"]],
                None,
                "Empty row not supported.",
            ),
            (
                None,
                [["a", "b"], ["c", "d", "e"], ["f", "g"]],
                None,
                "Mixed lengths in table content.",
            ),
            (
                ["h1", "h2", "h3"],
                [["a", "b"], ["c", "d"], ["e", "f"]],
                None,
                "Mixed lengths in table content.",
            ),
            (
                ["h1", "h2"],
                [["a", "b"], ["c", "d"], ["e", "f"]],
                [ContentAlignment.RIGHT],
                "'alignment' list should have length 2",
            ),
        ),
    )
    def test_validate_rejects_invalid_entries(
        self, headers, rows, alignment, expected_msg
    ):
        with pytest.raises(ValueError) as error:
            Table(headers=headers, rows=rows, alignment=alignment)

        assert expected_msg in str(error.value)

    def test_column_sizes(self):
        table = Table(
            ["header1", "h2", "h3", "h4"],
            [
                ["a", "bc", "de", "f"],
                ["b", "de", "fg", "wow this is a really big string of data"],
                ["c", "fg", "hijkl", "m"],
            ],
        )
        assert table.column_sizes == [7, 2, 5, 39]

    def test_line_length(self):
        table = Table(
            ["header1", "h2", "h3", "h4"],
            [
                ["a", "bc", "de", "f"],
                ["b", "de", "fg", "wow this is a really big string of data"],
                ["c", "fg", "hijkl", "m"],
            ],
        )
        assert table._get_line_length() == 59

    @mock.patch(M_PATH + "sys.stdout.isatty", return_value=True)
    def test_wrap_last_column(self, _m_is_tty):
        table = Table(
            ["header1", "h2", "h3", "h4"],
            [
                ["a", "bc", "de", "f"],
                ["b", "de", "fg", "wow this is a really big string of data"],
                ["c", "fg", "hijkl", "m"],
            ],
        )
        assert table.wrap_last_column(max_length=40) == [
            table.rows[0],
            ["b", "de", "fg", "wow this is a"],
            [" ", " ", " ", "really big string of"],
            [" ", " ", " ", "data"],
            table.rows[2],
        ]

    @mock.patch(M_PATH + "sys.stdout.isatty", return_value=True)
    def test_to_string_wraps_to_length(self, _m_is_tty, FakeConfig):
        POFC.init(FakeConfig())
        table = Table(
            ["header1", "h2", "h3", "h4"],
            [
                ["a", "bc", "de", "f"],
                ["b", "de", "fg", "wow this is a really big string of data"],
                ["c", "fg", "hijkl", "m"],
            ],
        )
        assert table.to_string(line_length=40) == textwrap.dedent(
            """\
            \x1b[1mheader1  h2  h3     h4\x1b[0m
            a        bc  de     f
            b        de  fg     wow this is a
                                really big string of
                                data
            c        fg  hijkl  m
            """
        )

    @mock.patch(M_PATH + "sys.stdout.isatty", return_value=False)
    def test_columns_align_to_the_right(self, _m_is_tty, FakeConfig):
        POFC.init(FakeConfig())
        table = Table(
            ["header1", "h2", "h3", "h4"],
            [
                ["a", "bc", "de", "f"],
                ["b", "de", "fg", "wow this is a really big string of data"],
                ["c", "fg", "hijkl", "m"],
            ],
            alignment=[ContentAlignment.RIGHT] * 4,
        )
        assert table.to_string() == textwrap.dedent(
            """\
            header1  h2     h3                                       h4
                  a  bc     de                                        f
                  b  de     fg  wow this is a really big string of data
                  c  fg  hijkl                                        m
            """  # noqa: E501
        )
        assert table.to_string(line_length=40) == textwrap.dedent(
            """\
            header1  h2     h3                    h4
                  a  bc     de                     f
                  b  de     fg         wow this is a
                                really big string of
                                                data
                  c  fg  hijkl                     m
            """
        )

    @mock.patch(M_PATH + "sys.stdout.isatty", return_value=True)
    @mock.patch(
        M_PATH + "os.get_terminal_size",
        return_value=mock.MagicMock(columns=40),
    )
    def test_to_string_wraps_to_terminal_size(
        self, _m_terminal_size, _m_is_tty, FakeConfig
    ):
        POFC.init(FakeConfig())
        table = Table(
            ["header1", "h2", "h3", "h4"],
            [
                [create_link(text="a", url="example.com"), "bc", "de", "f"],
                ["b", "de", "fg", "wow this is a really big string of data"],
                ["c", "fg", "hijkl", "m"],
            ],
        )
        assert table.to_string() == textwrap.dedent(
            """\
            \x1b[1mheader1  h2  h3     h4\x1b[0m
            \x1b]8;;example.com\x1b\\a\x1b]8;;\x1b\\        bc  de     f
            b        de  fg     wow this is a
                                really big string of
                                data
            c        fg  hijkl  m
            """
        )

    @mock.patch(M_PATH + "sys.stdout.isatty", return_value=False)
    # There is no terminal size if you are not in a tty
    @mock.patch(
        M_PATH + "os.get_terminal_size",
        side_effect=OSError(),
    )
    def test_to_string_no_wrap_if_no_tty(
        self, _m_terminal_size, _m_is_tty, FakeConfig
    ):
        POFC.init(FakeConfig())
        table = Table(
            ["header1", "h2", "h3", "h4"],
            [
                [create_link(text="a", url="example.com"), "bc", "de", "f"],
                [
                    "b",
                    "de",
                    "fg",
                    "wow this is a really big string of data" * 3,
                ],
                ["c", "fg", "hijkl", "m"],
            ],
        )
        assert table.to_string() == textwrap.dedent(
            """\
            header1  h2  h3     h4
            a        bc  de     f
            b        de  fg     wow this is a really big string of datawow this is a really big string of datawow this is a really big string of data
            c        fg  hijkl  m
            """  # noqa: E501
        )


class TestBlock:
    @mock.patch(M_PATH + "sys.stdout.isatty", return_value=False)
    @mock.patch(
        M_PATH + "os.get_terminal_size",
        side_effect=OSError(),
    )
    def test_indents_and_wraps_content_when_len_specified(
        self, _m_terminal_size, _m_is_tty, FakeConfig
    ):
        POFC.init(FakeConfig())
        block = Block(
            title="Example Title",
            content=[
                "Smaller content line",
                "A slightly bigger line which needs to be wrapped to fit the screen",  # noqa: E501
                Block(
                    title="Inner block",
                    content=[
                        "Another small content line",
                        "Another slightly bigger line which needs to be wrapped to fit the screen",  # noqa: E501
                    ],
                ),
                Table(
                    rows=[
                        [
                            "1",
                            "Table bigger last column which needs to be wrapped to fit the screen",  # noqa: E501
                        ],
                        [
                            "2",
                            "Table bigger last column which needs to be wrapped to fit the screen",  # noqa: E501
                        ],
                        [
                            "3",
                            "Table bigger last column which needs to be wrapped to fit the screen",  # noqa: E501
                        ],
                    ]
                ),
            ],
        )
        assert block.to_string() == textwrap.dedent(
            """\
            Example Title
                Smaller content line
                A slightly bigger line which needs to be wrapped to fit the screen
                Inner block
                    Another small content line
                    Another slightly bigger line which needs to be wrapped to fit the screen
                1  Table bigger last column which needs to be wrapped to fit the screen
                2  Table bigger last column which needs to be wrapped to fit the screen
                3  Table bigger last column which needs to be wrapped to fit the screen
            """  # noqa: E501
        )
        assert block.to_string(line_length=40) == textwrap.dedent(
            """\
            Example Title
                Smaller content line
                A slightly bigger line which needs
                to be wrapped to fit the screen
                Inner block
                    Another small content line
                    Another slightly bigger line
                    which needs to be wrapped to fit
                    the screen
                1  Table bigger last column which
                   needs to be wrapped to fit the
                   screen
                2  Table bigger last column which
                   needs to be wrapped to fit the
                   screen
                3  Table bigger last column which
                   needs to be wrapped to fit the
                   screen
            """
        )



import json
from urllib.parse import urlencode

import mock
import pytest

from uaclient import http
from uaclient.http.serviceclient import UAServiceClient


class OurServiceClient(UAServiceClient):
    @property
    def cfg_url_base_attr(self):
        return "contract_url"


class TestRequestUrl:

    # TODO: Non error-path tests

    @pytest.mark.parametrize(
        "m_kwargs", ({"a": 1, "b": "2", "c": "try me"}, {})
    )
    @mock.patch("uaclient.http.readurl")
    def test_url_query_params_append_querystring(
        self, m_readurl, m_kwargs, FakeConfig
    ):
        m_readurl.return_value = (m_kwargs, {})  # (response, resp_headers)

        cfg = FakeConfig()
        url = "http://example.com/"
        cfg.cfg["contract_url"] = url
        client = OurServiceClient(cfg=cfg)
        assert (m_kwargs, {}) == client.request_url("/", query_params=m_kwargs)
        if m_kwargs:
            url += "?" + urlencode(m_kwargs)
        assert [
            mock.call(
                url=url,
                data=None,
                headers=client.headers(),
                method=None,
                timeout=30,
                log_response_body=True,
            )
        ] == m_readurl.call_args_list


URL_FAKES = {
    "http://a": [{"code": 200, "response": {"key": "val", "key2": "val2"}}],
    "http://anerror": [{"code": 404, "response": "nothing to see"}],
    "http://multiresp": [
        {"code": 200, "response": "resp1"},
        {
            "code": 200,
            "response": "resp2",
            "headers": {"content-type": "application/json"},
        },
    ],
}


class Test_GetResponseOverlay:
    @pytest.mark.parametrize(
        "url,expected",
        (
            ("http://a", URL_FAKES["http://a"]),
            ("http://multiresp", URL_FAKES["http://multiresp"]),
        ),
    )
    def test_overlay_returns_matching_url(
        self, url, expected, FakeConfig, tmpdir
    ):
        overlay_path = tmpdir.join("overlay.json")
        overlay_path.write(json.dumps(URL_FAKES))
        cfg = FakeConfig()
        cfg.override_features(
            {"serviceclient_url_responses": overlay_path.strpath}
        )
        client = OurServiceClient(cfg=cfg)
        assert expected == client._get_response_overlay(url)
        # url overrides are cached and not re-read from disk
        overlay_path.write(json.dumps({}))
        assert expected == client._get_response_overlay(url)


class Test_GetFakeResponses:
    @pytest.mark.parametrize(
        "url,overlay,responses",
        (
            # When URL has no fakes
            ("http://a", {}, []),
            # When URL has 1 fake, repeat that response for all calls
            (
                "http://a",
                URL_FAKES,
                [URL_FAKES["http://a"][0]] * 2,
            ),
            # When URL has >1 fake, pop through that list and repeat last item
            (
                "http://multiresp",
                URL_FAKES,
                [URL_FAKES["http://multiresp"][0]]
                + [
                    URL_FAKES["http://multiresp"][1],
                    URL_FAKES["http://multiresp"][1],
                ]
                * 2,
            ),
            # When URL fake is code != 200, return that code
            (
                "http://anerror",
                URL_FAKES,
                [URL_FAKES["http://anerror"][0]],
            ),
        ),
    )
    def test_overlay_returns_matching_url(
        self, url, overlay, responses, FakeConfig, tmpdir
    ):
        overlay_path = tmpdir.join("overlay.json")
        overlay_path.write(json.dumps(overlay))
        cfg = FakeConfig()
        cfg.override_features(
            {"serviceclient_url_responses": overlay_path.strpath}
        )
        client = OurServiceClient(cfg=cfg)
        for response in responses:
            assert http.HTTPResponse(
                code=response["code"],
                headers=response.get("headers", {}),
                body=json.dumps(response["response"], sort_keys=True),
                json_dict=(
                    response["response"]
                    if isinstance(response["response"], dict)
                    else {}
                ),
                json_list=(
                    response["response"]
                    if isinstance(response["response"], list)
                    else []
                ),
            ) == client._get_fake_responses(url)



from typing import Optional

from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.data_types import DataObject, Field, StringDataValue
from uaclient.files.state_files import apt_news_raw_file


class CurrentNewsResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "current_news",
            StringDataValue,
            required=False,
            doc=(
                "The current APT News to be displayed for the system. This"
                " could be a str with up to three lines (i.e. up to two"
                " ``\\n`` characters). If there is no APT News to be"
                " displayed, this will be ``None``."
            ),
        ),
    ]

    def __init__(self, *, current_news: Optional[str]):
        self.current_news = current_news


def current_news() -> CurrentNewsResult:
    return _current_news(UAConfig())


def _current_news(cfg: UAConfig) -> CurrentNewsResult:
    """
    This endpoint returns the current APT News that gets displayed in
    `apt upgrade`.
    """
    return CurrentNewsResult(current_news=apt_news_raw_file.read())


endpoint = APIEndpoint(
    version="v1",
    name="CurrentNews",
    fn=_current_news,
    options_cls=None,
)

_doc = {
    "introduced_in": "29",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.apt_news.current_news.v1 import current_news

result = current_news().current_news
""",
    "result_class": CurrentNewsResult,
    "exceptions": [],
    "example_cli": "pro api u.apt_news.current_news.v1",
    "example_json": """
{
    "current_news":"This is a news message.\\nThis is the second line of the message.\\nAnd this is the third line."
}
""",  # noqa: E501
}
